/*
** Copyright (c) 2002 D. Richard Hipp
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** General Public License for more details.
** 
** You should have received a copy of the GNU General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA  02111-1307, USA.
**
** Author contact information:
**   drh@hwaci.com
**   http://www.hwaci.com/drh/
**
*******************************************************************************
**
** This file contains code used to browse through the CVS repository.
*/
#include "config.h"
#include "browse.h"
#include <sys/times.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <time.h>
#include <string.h>

/*
** This routine generates an HTML page that describes the complete
** revision history for a single file.
*/
static void revision_history(const char *zName, int showMilestones){
  char **az;
  int i;
  const char *zTail;

  if( zName[0]=='/' ) zName++;  /* Be nice to TortoiceCVS */
  zTail = strrchr(zName, '/');
  if( zTail ) zTail++;

  /* @ <h2>History of /%h(zName)</h2> */
  if( showMilestones ){
    common_add_action_item(mprintf("rlog?f=%t",zName), "Omit Milestones");
    az = db_query("SELECT filechng.cn, date, vers, nins, ndel, prevvers,"
                  "       message, user, branch "
                  "FROM filechng, chng "
                  "WHERE filename='%q' AND filechng.cn=chng.cn "
                  "UNION ALL "
                  "SELECT '',date,cn,NULL,NULL,NULL,message,user,branch "
                  "FROM chng "
                  "WHERE milestone=1 "
                  "ORDER BY 2 DESC", zName);
  } else {
    common_add_action_item(mprintf("rlog?f=%t&sms=1",zName), "Show Milestones");
    az = db_query("SELECT filechng.cn, date, vers, nins, ndel, prevvers,"
                  "       message, user, branch "
                  "FROM filechng, chng "
                  "WHERE filename='%q' AND filechng.cn=chng.cn "
                  "ORDER BY date DESC", zName);
  }

  common_header("History for /%h", zName);

  @ <table cellpadding=0 cellspacing=0 border=0>
  for(i=0; az[i]; i+=9){
    time_t t;
    struct tm *pTm;
    char zDate[100];

    t = atoi(az[i+1]);
    pTm = localtime(&t);
    strftime(zDate, sizeof(zDate), "%Y-%b-%d %H:%M", pTm);
    if( i==0 ){
      @ <thread><tr><th>Date</th><th width=80>Version</th>
      @         <th>Description</th></tr>
      @ <tbody>
    }
    if( i%2 ){
      @ <tr bgcolor="%s(BG4)" class="bkgnd4">
    }else{
      @ <tr>
    }
    @ <td align="right" valign="top"><nobr>%s(zDate)</nobr></td>
    if( az[i][0]==0 ){
      @ <td align="center" valign="top">
      common_icon("box");
      @ </td>
      if( az[i+8] && az[i+8][0] ){
        @ <td align="left" bgcolor="%s(BG5)" class="bkgnd5">
        @ Milestone
        output_chng(atoi(az[i+2]));
        @    on branch %h(az[i+8]):
      }else{
        @ <td align="left">
        @ Milestone
        output_chng(atoi(az[i+2]));
      }
    }else{
      @ <td valign="top" align="center">&nbsp;&nbsp;
      @ <a href="fileview?f=%T(zName)&v=%T(az[i+2])">
      @    %h(printable_vers(az[i+2]))</a>
      @ &nbsp;&nbsp;</td>
      if( az[i+8] && az[i+8][0] ){
        @ <td bgcolor="%s(BG5)" class="bkgnd5">Check-in
        output_chng(atoi(az[i]));
        @     on branch %h(az[i+8]):
      }else{
        @ <td>Check-in
        output_chng(atoi(az[i]));
        @ :
      }
    }
    output_formatted(az[i+6], 0);
    @&nbsp;By %h(az[i+7]).
    if( az[i][0]!=0 ){ /* Can't diff a Milestone */
      if( g.okCheckout && az[i+5] && az[i+5][0] ){
        @ <a href="filediff?f=%T(zName)&v1=%T(az[i+5])&v2=%T(az[i+2])">
        @ (diff)</a>
      }
    }
    @ </td></tr>
  }
  if( i==0 ){
    @ <tr><td>Nothing is known about this file</td></tr>
  }
  @ </table>
}

/*
** WEBPAGE: /rlog
**
** This page lists the revision history for a single file.   Hyperlinks
** allow the file to be diffed or annotated.
*/
void browse_rlog(void){
  char *zDir, *z;
  int showMilestones;
  const char *zFile;

  login_check_credentials();
  if( !g.okCheckout ){ login_needed(); return; }
  throttle(1,0);
  common_standard_menu(0, "search?f=1");
  showMilestones = atoi(PD("sms","0"));
  history_update(0);
  zFile = PD("f","");
  /* Make sure we always have '/' in zFile, otherwise link to parent
  ** directory won't work for file in repository root.
  */
  if( strrchr(zFile, '/') ){
    zDir = mprintf("%T?d=%T", default_browse_url(), zFile);
  }else{
    zDir = mprintf("%T?d=/%T", default_browse_url(), zFile);
  }
  z = strrchr(zDir, '/' );
  if( z ){ *z = 0;}
  common_add_action_item(zDir, "Directory");
  revision_history(zFile, showMilestones);
  common_footer();
}

/*
** WEBPAGE: /filediff
**
** Show the differences between two versions of a file
*/
void browse_filediff(void){
  const char *zFile = P("f");
  const char *zV1 = P("v1");
  const char *zV2 = P("v2");

  login_check_credentials();
  if( !g.okCheckout ){ login_needed(); return; }
  throttle(1,0);
  if( zFile==0 || zV1==0 || zV2==0 ){ cgi_redirect("index"); return; }
  common_standard_menu(0, "search?f=1");
  common_add_action_item(mprintf("rlog?f=%T", zFile), "History");
  common_header("Difference in %h versions %h and %h", zFile,
      printable_vers(zV1), printable_vers(zV2));
  if( diff_versions(zV1, zV2, zFile) ){
    @ <b>Diff failed!</b>
  }
  common_footer();
}

/*
** WEBPAGE: /dir
**
** List all of the repository files in a single directory.
*/
void browse_dir(void){
  const char *zName;
  char *zDir;
  char *zBase;
  char **az;
  int i, j;
  int n;
  int nRow;
  const char *zRoot;
  const char *zCookieName;
  int nCookieLife;

  login_check_credentials();
  if( !g.okCheckout ){ login_needed(); return; }
  throttle(1,0);
  history_update(0);
  common_standard_menu("dir", "search?f=1");
  /* P("sc") is set only when user explicitly switches to Long/Short view,
  ** via action bar link. In that case we make that users preference
  ** persistent via cookie.
  */
  if( P("sc") ){
    zCookieName = mprintf("%t_browse_url",g.zName);
    nCookieLife = 86400*atoi(db_config("browse_url_cookie_life","90"));
    if( nCookieLife ){
      cgi_set_cookie(zCookieName, "dir", 0, nCookieLife);
    }
  }
  zName = PD("d","");
  if( zName==0 ){
    zName = "";
  }
  if( zName[0] ){
    common_add_action_item(
      mprintf("timeline?x=1&c=2&dm=1&px=%h",zName),
      "Activity"
    );
  }
  zDir = mprintf("%s", zName);
  zBase = strrchr(zDir, '/');
  if( zBase==0 ){
    zBase = zDir;
    zDir = "";
  }else{
    *zBase = 0;
    zBase++;
  }
  if( zName && zName[0] ){
    /* this looks like navigation, but it's relative to the current page
    */
    common_add_action_item("dir", "Top");
    common_add_action_item(mprintf("dir?d=%T",zDir), "Up");
    common_add_action_item(mprintf("dirview?d=%T&sc=1",zName), "Long");
  }else{
    common_add_action_item("dirview?sc=1","Long");
  }
  az = db_query("SELECT base, isdir FROM file WHERE dir='%q' ORDER BY base",
                 zName);
  for(n=0; az[n*2]; n++){}
  if( zName[0] ) n++;
  nRow = (n+3)/4;
  if( zName[0] ){ zName = mprintf("%s/",zName); }
  zRoot = db_config("cvsroot","");
  common_header("Directory /%h", zName);
  /* @ <h2>Contents of directory /%h(zName)</h2> */
  @ <table width="100%%">
  @ <tr>
  for(i=j=0; i<4; i++){
    @ <td valign="top" width="25%%">
    n = 0;
    if( i==0 && zName[0] ){
      @ <a href="dir?d=%T(zDir)">
      common_icon("backup");
      @ </a>&nbsp;<a href="dir?d=%T(zDir)">..</a><br>
      n++;
    }
    while( n<nRow && az[j] ){
      if( atoi(az[j+1]) ){
        @ <a href="dir?d=%T(zName)%T(az[j])">
        common_icon("dir");
        @ </a>&nbsp;<a href="dir?d=%T(zName)%T(az[j])">%h(az[j])/</a><br>
      }else{
        char *zIcon;
        char *zFilename = mprintf("%s%s", zName, az[j]);
        if(is_file_available(zFilename)){
          zIcon = "file";
        }else{
          zIcon = "del";
        }
        if( zFilename!=0 ) free(zFilename);
        @ <a href="rlog?f=%T(zName)%T(az[j])">
        common_icon(zIcon);
        @ </a>&nbsp;<a href="rlog?f=%T(zName)%T(az[j])">%h(az[j])</a><br>
      }
      n++;
      j += 2;
    }
    @ </td>
  }
  @ </tr></table>
  common_footer();
}


/*
** This routine is used to represent age of files in english text.
** For example "1 week", "3 days", etc.
** It takes integer representing unix time of file's last modification and
** calculates it's age relative to current time.
*/
static char *file_age_to_text(int nModified){
  int nAge, n;
  int nYear  = 31536000; /* Number of seconds in a year */
  int nMonth =  2592000; /* Number of seconds in a month */
  int nWeek  =   604800; /* Number of seconds in a week */
  int nDay   =    86400; /* Number of seconds in a day */
  
  if( nModified<=0 ){
    /* FIXME: some error handling would be nice here */
    return NULL;
  }
  
  nAge = (int)time(0)-nModified;
  if( nAge<0 ){
    /* FIXME: some error handling would be nice here */
    return NULL;
  }
  
  if( (n = nAge/nYear)>1 ){
    return mprintf("%d years", n);
  }else if( (n = nAge/nMonth)>1 ){
    return mprintf("%d months", n);
  }else if( (n = nAge/nWeek)>1 ){
    return mprintf("%d weeks", n);
  }else if( (n = nAge/nDay)>1 ){
    return mprintf("%d days", n);
  }else if( (n = nAge/3600)>1 ){
    return mprintf("%d hours", n);
  }else{
    n = nAge/60;
    if( n<=1 ){
      return mprintf("1 minute");
    }else{
      return mprintf("%d minutes", n);
    }
  }
}

/*
** WEBPAGE: /dirview
**
** This is a "long view" version of /dir page.
** List all of the repository files in a single directory and display 
** information about their last change.
*/
void browse_dirview(void){
  const char *zName;
  char *zDir;
  char *zBase;
  char **az;
  char **azChng;
  int i;
  const char *zRoot;
  const char *zCookieName;
  int nCookieLife;

  login_check_credentials();
  if( !g.okCheckout ){ login_needed(); return; }
  throttle(1,0);
  history_update(0);
  common_standard_menu("dirview", "search?f=1");
  /* P("sc") is set only when user explicitly switches to Long/Short view,
  ** via action bar link. In that case we make that users preference
  ** persistent via cookie.
  */
  if( P("sc") ){
    zCookieName = mprintf("%t_browse_url",g.zName);
    nCookieLife = 86400*atoi(db_config("browse_url_cookie_life","90"));
    if( nCookieLife ){
      cgi_set_cookie(zCookieName, "dirview", 0, nCookieLife);
    }
  }
  zName = PD("d","");
  if( zName==0 ){
    zName = "";
  }
  if( zName[0] ){
    common_add_action_item(
      mprintf("timeline?x=1&c=2&dm=1&px=%T",zName),
      "Activity"
    );
  }
  zDir = mprintf("%s", zName);
  zBase = strrchr(zDir, '/');
  if( zBase==0 ){
    zBase = zDir;
    zDir = "";
  }else{
    *zBase = 0;
    zBase++;
  }
  if( zName && zName[0] ){
    /* this looks like navigation, but it's relative to the current page
    */
    common_add_action_item(mprintf("dirview"), "Top");
    common_add_action_item(mprintf("dirview?d=%T",zDir), "Up");
    common_add_action_item(mprintf("dir?d=%T&sc=1",zName), "Short");
  }else{
    common_add_action_item("dir?sc=1", "Short");
  }
  az = db_query("SELECT base, isdir FROM file WHERE dir='%q' "
                "ORDER BY isdir DESC, base", zName);
  if( zName[0] ){ zName = mprintf("%s/",zName); }
  zRoot = db_config("cvsroot","");
  common_header("Directory /%h", zName);
  @ <table width="100%%" border=0 cellspacing=0 cellpadding=3>
  @ <tr>
  @ <th align="left">File</th>
  @ <th align="left">Vers</th>
  @ <th align="left">By</th>
  @ <th align="left">Age</th>
  @ <th align="left">Check-in</th>
  @ </tr>
  
  if( zName[0] ){
    @ <tr><td colspan="5">
    @ <a href="dirview?d=%T(zDir)">
    common_icon("backup");
    @ </a>&nbsp;<a href="dirview?d=%T(zDir)">..</a></td></tr>
  }
  for(i=0; az[i]; i+=2){
    if( (i/2)%2==0 ){
      @ <tr bgcolor="%s(BG4)" class="bkgnd4">
    }else{
      @ <tr>
    }
    
    if( atoi(az[i+1])==1 ){
      char *z = db_short_query(
        "SELECT max(chng.date) FROM filechng, chng WHERE "
        "      filechng.filename LIKE '%q%q%%' AND filechng.cn=chng.cn;",
        zName, az[i]
      );

      @ <td colspan="3">
      @ <a href="dirview?d=%T(zName)%T(az[i])">
      common_icon("dir");
      @ </a>&nbsp;<a href="dirview?d=%T(zName)%T(az[i])">%h(az[i])/</a></td>
      @ <td valign="middle" width="10%%">%h(file_age_to_text(atoi(z)))</td>
      @ <td></td>

      if(z) free(z);
    }else{
      char *zFilename = mprintf("%s%s", zName, az[i]);
      azChng = db_query(
        "SELECT fc.filename, fc.vers, fc.chngtype, c.user, c.date,"
        "       c.message, c.cn "
        "FROM (SELECT filename, MAX(cn) AS maxcn FROM filechng "
        "      WHERE filename='%q' GROUP BY filename) maxcn, "
        "      filechng fc, chng c "
        "WHERE fc.filename=maxcn.filename "
        "  AND fc.cn=maxcn.maxcn "
        "  AND fc.cn=c.cn;",
        zFilename
      );
      if( !azChng || !azChng[0] ){
        @ <h1>ERROR: No chng data for %h(zFilename)</h1>
        exit(1);
      }
      if( zFilename!=0 ) free(zFilename);
      
      @ <td valign="middle" width="30%%">
      @ <a href="rlog?f=%T(zName)%T(az[i])">
      if( atoi(azChng[2])==2 ){
        common_icon("del");
      }else{
        common_icon("file");
      }
      @ </a>&nbsp;<a href="rlog?f=%T(zName)%T(az[i])">%h(az[i])</a></td>
      @ <td valign="middle" width="5%%">
      @ <a href="fileview?f=%T(zName)%T(az[i])&v=%T(azChng[1])">
      @ %h(printable_vers(azChng[1]))</a></td> 
      @ <td valign="middle" width="8%%">%h(azChng[3])</td>
      @ <td valign="middle" width="8%%">%h(file_age_to_text( atoi(azChng[4]) ))</td>
      @ <td valign="middle">

      /* For svntrac, the cn is the same as the file revision. For everything
      ** else... However, even for svntrac it's useful to have a link to the
      ** /chngview rather than just the /fileview.
      */
      output_chng(atoi(azChng[6]));
      cgi_printf(" ");
      if( output_trim_message(azChng[5], MN_CKIN_MSG, MX_CKIN_MSG) ){
        output_formatted(azChng[5], 0);
        @ &nbsp;[...]
      }else{
        output_formatted(azChng[5], 0);
      }
      @ </td>
      db_query_free(azChng);
    }
    @</tr>
  }
  @ </table>
  db_query_free(az);
  common_footer();
}

/*
** WEBPAGE: /fileview
**
** Show the file in a HTML page. In the case of things like images, show the
** content embedded in the page.
*/
void browse_fileview(void){
  const char *zFile = g.zExtra ? g.zExtra : P("f");
  const char *zVers = PD("v","");
  char *zGetFile;
  char *zDir, *z;
  char *zSuffix;
  char *zMime = "text/plain";  /* The default MIME type */

  /* The following table lists some alternative MIME types based on
  ** the file suffix
  */
  static const struct {
    char *zSuffix;
    char *zMime;
  } aMime[] = {
    { "html",  "text/html" },
    { "htm",   "text/html" },
    { "gif",   "image/gif" },
    { "jpeg",  "image/jpeg" },
    { "jpg",   "image/jpeg" },
    { "png",   "image/png" },
    { "pdf",   "application/pdf" },
    { "ps",    "application/postscript" },
    { "eps",   "application/postscript" },
  };

  login_check_credentials();
  if( !g.okCheckout ){ login_needed(); return; }
  throttle(1,0);
  common_standard_menu(0, "search?f=1");
  history_update(0);

  /* Make sure we always have '/' in zFile, otherwise link to parent
  ** directory won't work for file in repository root.
  */
  if( strrchr(zFile, '/') ){
    zDir = mprintf("%T?d=%T", default_browse_url(), zFile);
  }else{
    zDir = mprintf("%T?d=/%T", default_browse_url(), zFile);
  }
  z = strrchr(zDir, '/' );
  if( z ){ *z = 0;}
  common_add_nav_item(zDir, "Directory");

  zGetFile = mprintf("getfile?f=%T&v=%T", zFile, zVers);
  common_add_action_item(zGetFile, "Raw");

  common_header("%h %h", zFile, printable_vers(zVers));

  /* sort out the MIME type. We output HTML, but some things are embeddable. */
  zSuffix = strrchr(zFile, '.');
  if( zSuffix ){
    char zLine[2000];
    int i;
    zSuffix++;
    for(i=0; zSuffix[i] && i<sizeof(zLine)-1; i++){
      zLine[i] = tolower(zSuffix[i]);
    }
    zLine[i] = 0;
    for(i=0; i<sizeof(aMime)/sizeof(aMime[0]); i++){
      if( strcmp(zLine, aMime[i].zSuffix)==0 ){
        zMime = aMime[i].zMime;
        break;
      }
    }
  }

  @ <a href="rlog?f=%T(zFile)">%h(zFile)</a>
  @ <a href="%s(zGetFile)">%h(zVers)</a><hr>

  /* For image types, embed in the page. Anything else, try to inline */
  if( !strncmp(zMime,"image/",6) ){
    @ <img src="%s(zGetFile)" alt="%h(zFile) %h(zVers)">
  }else{
    if( dump_version(zVers,zFile,0) ){
      cgi_redirect("index");
      return;
    }
  }

  @ <hr>
  common_footer();
}

/*
** WEBPAGE: /getfile
**
** Return the complete content of a file
*/
void browse_getfile(void){
  const char *zFile = g.zExtra ? g.zExtra : P("f");
  const char *zVers = P("v");
  char *zSuffix;
  char *zMime = "text/plain";  /* The default MIME type */

  /* The following table lists some alternative MIME types based on
  ** the file suffix
  */
  static const struct {
    char *zSuffix;
    char *zMime;
  } aMime[] = {
    { "html",  "text/html" },
    { "htm",   "text/html" },
    { "gif",   "image/gif" },
    { "jpeg",  "image/jpeg" },
    { "jpg",   "image/jpeg" },
    { "png",   "image/png" },
    { "pdf",   "application/pdf" },
    { "ps",    "application/postscript" },
    { "eps",   "application/postscript" },
  };

  login_check_credentials();
  if( !g.okCheckout || zFile==0 ){ login_needed(); return; }
  throttle(1,0);

  if( zVers!= 0 ){
    /* A database query is almost definitely going to be faster than having
    ** to pull from from the repository, so we might as well try this first.
    */
    char *z = db_short_query("SELECT chng.date FROM filechng, chng "
                             "WHERE filechng.filename='%q' "
                             "      AND filechng.vers='%q' "
                             "      AND filechng.cn=chng.cn ",
                             zFile, zVers);
    if( z ){
      cgi_modified_since(atoi(z));
      cgi_append_header(mprintf("Last-Modified: %h\r\n",
                        cgi_rfc822_datestamp(atoi(z))));
      free(z);
    }
  }

  if( dump_version(zVers,zFile,1) ){
    cgi_redirect("index");
    return;
  }

  /* sort out the MIME type */
  zSuffix = strrchr(zFile, '.');
  if( zSuffix ){
    char zLine[2000];
    int i;
    zSuffix++;
    for(i=0; zSuffix[i] && i<sizeof(zLine)-1; i++){
      zLine[i] = tolower(zSuffix[i]);
    }
    zLine[i] = 0;
    for(i=0; i<sizeof(aMime)/sizeof(aMime[0]); i++){
      if( strcmp(zLine, aMime[i].zSuffix)==0 ){
        zMime = aMime[i].zMime;
        break;
      }
    }
  }
  if( zVers && zVers[0] ){
    g.isConst = 1;
  }
  cgi_set_content_type(zMime);
  cgi_set_status(200, "OK");
  return;
}
